package analyzer

import "fmt"
import "github.com/alecthomas/participle/v2"
import "github.com/alecthomas/participle/v2/lexer"
import "git.tebibyte.media/sashakoshka/fspl/entity"

// analyzeMethod analyzes a method that is directly owned by the specified type.
func (this *Tree) analyzeMethod (
	pos lexer.Position,
	typeName string,
	name string,
) (
	*entity.Method,
	error,
) {
	// get parent typedef
	owner, err := this.analyzeTypedef(pos, typeName, false)
	if err != nil { return nil, err }
	
	// return if exists already
	if method, exists := owner.Methods[name]; exists {
		return method, nil
	}
	
	// error if method is missing
	method, exists := this.rawMethods[typeName + "." + name]
	if !exists {
		return nil, participle.Errorf (
			pos, "no method named %s.%s",
			typeName, name)
	}

	// create a new scope context for this function
	this.pushScopeContext(method)
	this.pushScope(method)
	defer this.popScopeContext()
	defer this.popScope()
	
	// analyze signature and add arguments to root scope of function
	method.Signature, err = this.assembleSignatureMap(method.Signature)
	if err != nil { return method, err }
	for name, argument := range method.Signature.ArgumentMap {
		argument.Ty, err = this.analyzeType(argument.Ty, false)
		this.addVariable(argument)
		method.Signature.ArgumentMap[name] = argument
		if err != nil { return method, err }
	}
	method.Signature.Return, err =
		this.analyzeType(method.Signature.Return, false)
	if err != nil { return method, err }

	// add owner to method
	method.This = &entity.Declaration {
		Pos:  method.Pos,
		Name: "this",
		Ty:   &entity.TypePointer {
			Pos:        method.Pos,
			Referenced: &entity.TypeNamed {
				Pos:  method.Pos,
				Name: typeName,
				Type: owner.Type,
			},
		},
	}
	this.addVariable(method.This)

	// add incomplete method to complete methods because there is enough
	// information for it to be complete from the point of view of other
	// parts of the code
	owner.Methods[name] = method

	// analyze method body
	if method.Body != nil {
		body, err := this.analyzeExpression (
			method.Signature.Return, strict, method.Body)
		if err != nil { return nil, err }
		method.Body = body
	}
	return method, nil
}

func (this *Tree) methodExists (typeName, name string) bool {
	_, exists := this.rawMethods[typeName + "." + name]
	return exists
}

// analyzeMethodOrBehavior returns *entity.Signature if it found an interface
// behavior, and *entity.Method if it found a method.
func (this *Tree) analyzeMethodOrBehavior (
	pos lexer.Position,
	ty entity.Type,
	name string,
) (
	any,
	error,
) {
	return this.analyzeMethodOrBehaviorInternal(pos, ty, name, true)
}

func (this *Tree) analyzeMethodOrBehaviorInternal (
	pos             lexer.Position,
	ty              entity.Type,
	name            string,
	pierceReference bool,
) (
	any,
	error,
) {
	switch ty.(type) {
	case *entity.TypeNamed:
		ty := ty.(*entity.TypeNamed)
		if this.methodExists(ty.Name, name) {
			method, err := this.analyzeMethod(pos, ty.Name, name)
			if err != nil { return nil, err }
			return method, nil
		} else {
			return this.analyzeMethodOrBehaviorInternal (
				pos, ty.Type, name, false)
		}

	case *entity.TypeInterface:
		ty := ty.(*entity.TypeInterface)
		if behavior, ok := ty.BehaviorMap[name]; ok {
			return behavior, nil
		} else {
			return nil, participle.Errorf (
				pos, "no behavior or method named %s",
				name)
		}

	case *entity.TypePointer:
		if pierceReference {
			ty := ty.(*entity.TypePointer)
			return this.analyzeMethodOrBehaviorInternal (
				pos, ty.Referenced, name, false)
		} else {
			return nil, participle.Errorf (
				pos, "no method named %s defined on this type",
				name)
		}
	
	case *entity.TypeSlice,
		*entity.TypeArray,
		*entity.TypeStruct,
		*entity.TypeInt,
		*entity.TypeFloat,
		*entity.TypeWord:
		
		return nil, participle.Errorf (
			pos, "no method named %s defined on this type",
			name)
		
	default: panic(fmt.Sprint("BUG: analyzer doesnt know about type ", ty))
	}
}
