package entity

import "fmt"
import "github.com/alecthomas/participle/v2/lexer"

// TopLevel is any construct that is placed at the root of a file.
type TopLevel interface {
	topLevel ()
}

// Access determines the external access rule for a top-level entity.
type Access int; const (
	AccessPrivate Access = iota
	AccessRestricted
	AccessPublic
)

func (this *Access) String () string {
	switch *this {
	case AccessPrivate:    return "-"
	case AccessRestricted: return "~"
	case AccessPublic:     return "+"
	default: return fmt.Sprintf("entity.Access(%d)", this)
	}
}

func (this *Access) Capture (values []string) error {
	switch values[0] {
	case "-": *this = AccessPrivate
	case "~": *this = AccessRestricted
	case "+": *this = AccessPublic
	}
	return nil
}

// Typedef binds a type to a global identifier.
type Typedef struct {
	// Syntax
	Pos lexer.Position
	Acc    *Access `parser:" @('-' | '~' | '+')? "`
	Name   string  `parser:" @TypeIdent          "`
	Type   Type    `parser:" ':' @@              "`

	// Semantics
	Methods map[string] *Method
}
func (*Typedef) topLevel(){}
func (this *Typedef) String () string {
	output := ""
	if this.Acc != nil {
		output += fmt.Sprint(this.Acc, " ")
	}
	output += fmt.Sprint(this.Name, ": ", this.Type)
	if this.Methods != nil {
		for _, method := range this.Methods {
			output += fmt.Sprint("\n", method)
		}
	}
	return output
}

// Function binds a global identifier and argument list to an expression which
// is evaluated each time the function is called. If no expression is specified,
// the function is marked as external. Functions have an argument list, where
// each argument is passed as a separate variable. They return one value. All of
// these are typed.
type Function struct {
	// Syntax
	Pos lexer.Position
	Acc        *Access    `parser:" @('-' | '~' | '+')? "`
	Signature  *Signature `parser:" @@          "`
	LinkName   string     `parser:" @String?    "`
	Body       Expression `parser:" ( '=' @@ )? "`

	// Semantics
	Scope
}
func (*Function) topLevel(){}
func (this *Function) String () string {
	output := ""
	if this.Acc != nil {
		output += fmt.Sprint(this.Acc, " ")
	}
	output += this.Signature.String()
	if this.LinkName != "" {
		output += fmt.Sprint(" '", this.LinkName, "'")
	}
	if this.Body != nil {
		output += fmt.Sprint(" = ", this.Body)
	}
	return output
}

// Method is like a function, except localized to a defined type. Methods are
// called on an instance of that type, and receive a pointer to that instance
// via the "this" variable when they are run. Method names are not globally
// unique, but are unique within the type they are defined on.
type Method struct {
	// Syntax
	Pos lexer.Position
	Acc        *Access    `parser:" @('-' | '~' | '+')? "`
	TypeName   string     `parser:" @TypeIdent  "`
	Signature  *Signature `parser:" '.' @@      "`
	LinkName   string     `parser:" @String?    "`
	Body       Expression `parser:" ( '=' @@ )? "`

	// Semantics
	Scope
	Type Type
	This *Declaration
}
func (*Method) topLevel(){}
func (this *Method) String () string {
	output := ""
	if this.Acc != nil {
		output += fmt.Sprint(this.Acc, " ")
	}
	output += fmt.Sprint(this.TypeName, ".", this.Signature)
	if this.LinkName != "" {
		output += fmt.Sprint(" '", this.LinkName, "'")
	}
	if this.Body != nil {
		output += fmt.Sprint(" = ", this.Body)
	}
	return output
}
